"""
Defines the FITS Header Data Units (HDUs) of the OIFITS format: primary
header, binary tables containing interferometric data and instrumental 
properties.
"""

from .array import *
from .wavelength import *
from .target import *
from .vis import *
from .vis2 import *
from .t3 import *
from .flux import *
from .corr import *
from .inspol import *
from .primary import *
